Unit OBinFile;
{ Gestion buffrise d'un fichier }
{ K.B. avril-novembre 1994 }

{
Utilis par OFText (fichier texte),
            OFPcx  (fichier PCX),
            OFHelp (fichier d'aide),
            OFZip  (fichier ZIP),
}

{
  Cette unit dfinit l'objet TBinFile qui permet d'utiliser un fichier en
lecture ou en criture  travers un buffer. Les oprations de lecture et
d'criture se font ainsi en mmoire, les accs au disque dur ou  la
disquette n'ayant lieu que lorsque le buffer est plein.

  ATTENTION : la taille du buffer est dfinie lors de l'initialisation de
l'objet; ELLE DOIT IMPERATIVEMENT ETRE SUPERIEURE AU DOUBLE DE LA PLUS
GRANDE VARIABLE UTILISEE.
}

{$IFDEF debug}
 {$A+,B-,D+,E-,F-,I-,L+,N-,R+,S+,V-,W+,X+}
{$ELSE}
 {$A+,B-,D-,E-,F-,I-,L-,N-,R-,S-,V-,W+,X+}
{$ENDIF}

INTERFACE

Uses Dos,OFile;

Const
{ taille maximale d'une variable -1 }
 MaxIndex=65518;   { taille maximale d'une variable -1 }

{ codes d'erreurs }
 erMemoire    = $101;

Type
  PDatas=^TDatas;        { tableau reprsentant le buffer du fichier }
  TDatas=array[0..MaxIndex] of byte;

  PBinFile=^TBinFile;
  TBinFile=object(TFile)
   DataBuf:PDatas;        { pointeur sur buffer }
   DataFileSize:LongInt;  { taille du fichier en tenant compte des changements }
   DataFilePosit:LongInt; { position courante dans le fichier }
   MaxBufSize:Word;       { taille maximale du buffer }
   BufSize:Word;          { taille du buffer }
   BufDeb:LongInt;        { dbut du buffer dans le fichier }
   WriteFlag:Boolean;     { drapeau indiquant la ncessit de sauver le
                            buffer sur disque }

   Constructor Init(NomDeFichier:PathStr; ABufSize:Word);
   { ouverture d'un fichier existant et attribution de mmoire
     au buffer }
   Constructor Create(NomDeFichier:PathStr; ABufSize:Word);
   { cration d'un fichier et attribution de mmoire au buffer }
   Destructor Done;virtual;
   { fermeture du fichier et libration mmoire }
   Function GetErrorMsg:String;virtual;
   { renvoie un ventuel message d'erreur }
   Procedure SetFilePosit(num:longint);
   { fixe la position de lecture/criture dans le fichier }
   Function GetFilePosit:LongInt;
   { renvoie la position actuelle dans le fichier }
   Procedure SetEndPosit;
   { fixe la position de lecture/criture  la fin du fichier }
   Function EndOfFile:Boolean;
   { indique la fin du fichier }
   Function ReadByte:byte;
   { lecture de l'octet num dans le fichier }
   Procedure ReadVar(Var V; len:word);
   { lecture d'une variable V de longueur len  la position courante dans
     le fichier }
   Procedure ReadBuf;
   { chargement du buffer }
   Procedure WriteByte(b:byte);
   { criture de l'octet b  la position courante dans le fichier }
   Procedure WriteVar(Var V; len:word);
   { criture d'une variable V de longueur len  la position courante dans
     le fichier }
   Procedure WriteBuf;
   { criture du buffer dans le fichier sur disque }
   Function Find(S:String):Boolean;
   { recherche la chaine S  partir de la position courante dans le fichier
     et fixe la position courante au dbut de l'emplacement trouv en cas
     de succs et  la fin du fichier en cas d'chec. }
   End;

IMPLEMENTATION

{ Objet TBinFile }

Constructor TBinFile.Init(NomDeFichier:PathStr;ABufSize:Word);
Begin
 TFile.Init(NomDeFichier,stOpen);
 {initialisation des champs}
 WriteFlag:=false;
 DataBuf:=nil;
 DataFileSize:=0;
 DataFilePosit:=0;
 BufDeb:=0;
 BufSize:=0;
 if ABufSize>MaxIndex then ABufSize:=MaxIndex;
 MaxBufSize:=ABufSize;
 if not IsValid
    then exit;
 { taille du fichier }
 DataFileSize:=FSize;
 {allocation de mmoire pour le buffer}
 if MaxAvail< MaxBufSize+1024
    then ErrorFlag:=erMemoire
    else begin
          GetMem(DataBuf,MaxBufSize);
          { initialisation du buffer }
          ReadBuf;
         end;
End;

Constructor TBinFile.Create(NomDeFichier:PathStr;ABufSize:Word);
Begin
 TFile.Init(NomDeFichier,stCreate);
 {initialisation des champs}
 WriteFlag:=false;
 DataBuf:=nil;
 DataFileSize:=0;
 DataFilePosit:=0;
 BufDeb:=0;
 BufSize:=0;
 if ABufSize>MaxIndex then ABufSize:=MaxIndex;
 MaxBufSize:=ABufSize;
 if not IsValid
    then exit;
 {allocation de mmoire pour le buffer}
 if MaxAvail< MaxBufSize+1024
    then ErrorFlag:=erMemoire
    else GetMem(DataBuf,MaxBufSize);
End;

Destructor TBinFile.Done;
Begin
 if IsValid
    then WriteBuf;
 if DataBuf<>nil
    then FreeMem(DataBuf,MaxBufSize);
 TFile.Done;
End;

Function TBinFile.GetErrorMsg:String;
Begin
 case ErrorFlag of
  erMemoire : GetErrorMsg:='Mmoire insuffisante pour crer le buffer'+chr(13)
                           +'du fichier '+FName+FExt;
  else GetErrorMsg:=TFile.GetErrorMsg;
  end;
End;

Procedure TBinFile.SetFilePosit(num:LongInt);
Begin
 if num<0 then num:=0;
 if num>DataFileSize
    then begin
          SetEndPosit;
          repeat
           WriteByte(32);
          until (DataFilePosit=num) or (ErrorFlag<>0);
         end
    else DataFilePosit:=num;
End;

Function TBinFile.GetFilePosit:LongInt;
Begin
 GetFilePosit:=DataFilePosit;
End;

Procedure TBinFile.SetEndPosit;
Begin
 DataFilePosit:=DataFileSize;
End;

Function TBinFile.EndOfFile:Boolean;
Begin
 EndOfFile:=DataFilePosit>=DataFileSize;
End;

Procedure TBinFile.ReadBuf;
Var avant:LongInt;
Begin
 avant:=MaxBufSize div 2;
 if DataFilePosit-avant<0
    then BufDeb:=0
    else BufDeb:=DataFilePosit-avant;
 FSeek(BufDeb);
 FReadC(DataBuf^[0],MaxBufSize,BufSize);
End;

Procedure TBinFile.WriteBuf;
Begin
 if WriteFlag
    then begin
          FSeek(BufDeb);
          FWrite(DataBuf^[0],BufSize);
         end;
 WriteFlag:=false;
End;

Function TBinFile.ReadByte:byte;
Begin
 ReadByte:=0;
 if EndOfFile then exit;
 if (DataFilePosit<BufDeb) or (DataFilePosit>=BufDeb+MaxBufSize)
    then begin
          WriteBuf;
          ReadBuf;
         end;
 ReadByte:=DataBuf^[DataFilePosit-BufDeb];
 SetFilePosit(DataFilePosit+1);
End;

Procedure TBinFile.ReadVar(Var V; len:word);
Var n:word;
Begin
 if EndOfFile then exit;
 if (DataFilePosit-BufDeb<0) or (DataFilePosit-BufDeb+len>=MaxBufSize)
    then begin
          WriteBuf;
          ReadBuf;
         end;
 Move(DataBuf^[DataFilePosit-BufDeb],V,len);
 SetFilePosit(DataFilePosit+len);
End;

Procedure TBinFile.WriteByte(b:byte);
Begin
 if (DataFilePosit<BufDeb) or (DataFilePosit-BufDeb>=MaxBufSize)
    then begin
          WriteBuf;
          ReadBuf;
         end;
 DataBuf^[DataFilePosit-BufDeb]:=b;
 WriteFlag:=true;
 if EndOfFile
    then begin
          inc(BufSize);
          inc(DataFileSize);
         end;
 SetFilePosit(DataFilePosit+1);
End;

Procedure TBinFile.WriteVar(Var V; len:word);
Begin
 if (DataFilePosit-BufDeb<0) or (DataFilePosit-BufDeb+len>=MaxBufSize)
    then begin
          WriteBuf;
          ReadBuf;
         end;
 Move(V,DataBuf^[DataFilePosit-BufDeb],len);
 WriteFlag:=true;
 if EndOfFile
    then begin
          inc(BufSize,len);
          inc(DataFileSize,len);
         end;
 SetFilePosit(DataFilePosit+len);
End;

Function TBinFile.Find(S:String):Boolean;
{ recherche d'une chaine de caractres en utilisant la fonction Pos de
  Turbo Pascal }
Var W:String;
    l,p:Byte;
    ok,fin:Boolean;
Begin
 ok:=false;
 fin:=false;
 repeat
  if DataFileSize-DataFilePosit>255
     then l:=255
     else l:=DataFileSize-DataFilePosit;
  W[0]:=chr(l);
  ReadVar(W[1],l);
  if EndOfFile
     then fin:=true;
  p:=Pos(S,W);
  if p<>0
     then begin
           ok:=true;
           SetFilePosit(DataFilePosit-l+p-1);
          end
     else SetFilePosit(DataFilePosit-length(S));
 until ok or fin or (not IsValid);
 Find:=ok;
End;

END.

{                        Fin du fichier OBinFile.Pas                        }
